/* ***************************************************************************
   *  Kindermann                                 WPK Produktentwicklung      *
   *                                             Schuljahr: 2010/11          *
   *                                                                         *
   *                            Pacman VGA 25MHz                             *
   *                        für das Universal-Board                          *
   *                                                                         *                                                                         *
   *    Dieses Programm basiert auf dem AVGA Project von Jaromir Dvorak      *
   *    (http://avga.prometheus4.com/) sowie der Pacman-Version von Albert   *
   *    Seward.                                                              *
   *                                                                         *
   *    Es kann unter den Bedingungen der GNU General Public License (GPL)   *
   *    weitergeben und/oder modifiziert werden.                             *
   *                                                                         *
   ***************************************************************************/


#ifndef _PACMAN_H_
#define _PACMAN_H_

#define APPLE                0x29
#define BANANA               0x2a
#define PEAR                 0x2b
#define CHERRY               0x2c
#define PACMAN_CLOSED_RIGHT  0x01
#define PACMAN_CLOSED_DOWN   0x02
#define PACMAN_CLOSED_LEFT   0x03
#define PACMAN_CLOSED_UP     0x04
#define PACMAN_OPEN_RIGHT    0x05
#define PACMAN_OPEN_DOWN     0x06
#define PACMAN_OPEN_LEFT     0x07
#define PACMAN_OPEN_UP       0x08

#define DOT0                 0x20
#define DOT1                 0x21
#define DOT2                 0x22
#define DOT3                 0x23

#define POINTS100            0x2e
#define POINTS100_GLOW       0x2f

#define BLINKY               0x25
#define CLYDE                0x26
#define PINKY                0x27
#define INKEY                0x28
#define EATABLE              0x24

#define EATEN                0x2d
#define HART                 0x1e

typedef struct {
	uint8_t x;
	uint8_t y;
	uint8_t ico;
#define DIR_RIGHT 0
#define DIR_LEFT  1
#define DIR_DOWN  2
#define DIR_UP    3
#define DIR_PAUSE 4
	unsigned char move;
	unsigned char dirw;
} character_t;

void pacman_start(void);
void pacman_wait_key(void);

#define PACMAN_LOGO "\011\012\013\014\015\016\017"

#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/io.h>

#define NCLOSED 0xff     /* Normally closed, (STK500) */
#define NOPEN   0x00     /* Normally open, (joystick) */
#define LED_MODE NCLOSED
#define KEY_PORT PORTC
#define KEY_PIN  PINC
#define KEY_DDR  DDRC
#define KEY_MODE    NCLOSED
#define KEY_RIGHT   0x10
#define KEY_LEFT    0x20
#define KEY_UP      0x04
#define KEY_DOWN    0x08
#define KEY_COIN    0x02
#define KEY_START   0x02
#define KEY_DIR (KEY_LEFT | KEY_RIGHT | KEY_UP | KEY_DOWN)
#define KEY_ANY (KEY_DIR | KEY_COIN | KEY_START)

void utoa10(uint16_t num, uint8_t *str);
void key_init(void);
uint8_t key_wait_timeout(uint8_t mask, uint8_t timeout);
uint8_t misc_rand(void);
#define EEPROM_ERROR      1
uint8_t boot_test(void);

#define key_read(mask)  ((KEY_PIN ^ KEY_MODE) & (mask))
#define delay(x) {uint8_t __i; for(__i = 0; __i < (x); __i++) sync(); }

#define PACMAN_BOARD { \
0x1b,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x16,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x1a, \
0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10, \
0x10,0x21,0x1b,0x1a,0x21,0x1b,0x11,0x1a,0x21,0x10,0x21,0x1b,0x11,0x1a,0x21,0x1b,0x1a,0x21,0x10, \
0x10,0x21,0x1d,0x1c,0x21,0x1d,0x11,0x1c,0x21,0x12,0x21,0x1d,0x11,0x1c,0x21,0x1d,0x1c,0x21,0x10, \
0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10, \
0x10,0x21,0x14,0x13,0x21,0x15,0x21,0x14,0x11,0x16,0x11,0x13,0x21,0x15,0x21,0x14,0x13,0x21,0x10, \
0x10,0x23,0x23,0x23,0x21,0x10,0x23,0x23,0x21,0x10,0x23,0x23,0x21,0x10,0x23,0x23,0x23,0x21,0x10, \
0x1d,0x11,0x11,0x1a,0x21,0x17,0x11,0x13,0x21,0x12,0x21,0x14,0x11,0x18,0x21,0x1b,0x11,0x11,0x1c, \
0x20,0x20,0x20,0x10,0x21,0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10,0x21,0x10,0x20,0x20,0x20, \
0x11,0x11,0x11,0x1c,0x21,0x12,0x21,0x1b,0x13,0x20,0x14,0x1a,0x21,0x12,0x21,0x1d,0x11,0x11,0x11, \
0x20,0x20,0x20,0x23,0x23,0x23,0x21,0x10,0x20,0x20,0x20,0x10,0x23,0x23,0x23,0x23,0x20,0x20,0x20, \
0x11,0x11,0x11,0x1a,0x21,0x15,0x21,0x1d,0x11,0x11,0x11,0x1c,0x21,0x15,0x21,0x1b,0x11,0x11,0x11, \
0x20,0x20,0x20,0x10,0x21,0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10,0x21,0x10,0x20,0x20,0x20, \
/* 0x11,0x11,0x11,0x1a,0x21,0x15,0x21,0x1d,0x20,0x20,0x20,0x1c,0x21,0x15,0x21,0x1b,0x11,0x11,0x11, */ \
/* 0x20,0x20,0x20,0x10,0x21,0x10,0x23,0x1c,0x23,0x23,0x23,0x1c,0x21,0x10,0x21,0x10,0x20,0x20,0x20, */ \
0x1b,0x11,0x11,0x1c,0x21,0x12,0x21,0x14,0x11,0x16,0x11,0x13,0x21,0x12,0x21,0x1d,0x11,0x11,0x1a, \
0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10, \
0x10,0x21,0x14,0x1a,0x21,0x14,0x11,0x13,0x21,0x12,0x21,0x14,0x11,0x13,0x21,0x1b,0x13,0x21,0x10, \
0x10,0x23,0x21,0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10,0x23,0x21,0x10, \
0x17,0x13,0x21,0x12,0x21,0x15,0x21,0x14,0x11,0x16,0x11,0x13,0x21,0x15,0x21,0x12,0x21,0x14,0x18, \
0x10,0x23,0x23,0x23,0x21,0x10,0x23,0x23,0x21,0x10,0x23,0x23,0x21,0x10,0x23,0x23,0x23,0x21,0x10, \
0x10,0x21,0x14,0x11,0x11,0x19,0x11,0x13,0x21,0x12,0x21,0x14,0x11,0x19,0x11,0x11,0x13,0x21,0x10, \
0x10,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x23,0x21,0x10, \
0x1d,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x1c, \
}

#endif /* !_PACMAN_H_ */
